﻿//Josip Medved <jmedved@jmedved.com>  http://www.jmedved.com

//2008-05-01: Initial version.
//2008-07-06: Changing properties to older C# form (so it can work in VS 2005).


using System;
using System.Threading;
using System.Windows.Forms;

namespace Medo.ComponentModel {

	/// <summary>
	/// Executes an operation on a separate thread.
	/// </summary>
	public class BackgroundWorker : System.ComponentModel.Component {

		private readonly object _syncRoot = new object();
		private readonly Control _control;

		/// <summary>
		/// Initializes new instance.
		/// </summary>
		public BackgroundWorker() {
			_control = new Control();
			IntPtr dummyHandle = _control.Handle;
		}


		/// <summary>
		/// Requests cancellation of a pending background operation.
		/// </summary>
		/// <exception cref="System.InvalidOperationException">WorkerSupportsCancellation is false.</exception>
		public void CancelAsync() {
			this.CancellationPending = true;
		}

		/// <summary>
		/// Raises the DoWork event.
		/// </summary>
		/// <param name="e">An System.EventArgs that contains the event data.</param>
		protected virtual void OnDoWork(DoWorkEventArgs e) {
			System.EventHandler<DoWorkEventArgs> eh = this.DoWork;
			if (eh != null) {
				eh(this, e);
			}
		}

		/// <summary>
		/// Raises the BackgroundWorker.ProgressChanged event.
		/// </summary>
		/// <param name="e">An System.EventArgs that contains the event data.</param>
		protected virtual void OnProgressChanged(ProgressChangedEventArgs e) {
			EventHandler<ProgressChangedEventArgs> eh = this.ProgressChanged;
			if (eh != null) {
				eh(this, e);
			}
		}

		/// <summary>
		/// Raises the BackgroundWorker.RunWorkerCompleted event.
		/// </summary>
		/// <param name="e">An System.EventArgs that contains the event data.</param>
		protected virtual void OnRunWorkerCompleted(RunWorkerCompletedEventArgs e) {
			System.EventHandler<RunWorkerCompletedEventArgs> eh = this.RunWorkerCompleted;
			if (eh != null) {
				eh(this, e);
			}
		}

		/// <summary>
		/// Raises the BackgroundWorker.ProgressChanged event.
		/// </summary>
		/// <param name="percentProgress">The percentage, from 0 to 100, of the background operation that is complete.</param>
		/// <param name="userState">The state object.</param>
		/// <exception cref="System.InvalidOperationException">The WorkerReportsProgress property is set to false.</exception>
		public void ReportProgress(int percentProgress, object userState) {
			if (this.WorkerReportsProgress == false) { throw new InvalidOperationException("The WorkerReportsProgress property is set to false."); }
			_control.Invoke(new OnProgressChangedDelegate(OnProgressChanged), new ProgressChangedEventArgs(percentProgress, userState));
		}

		/// <summary>
		/// Raises the BackgroundWorker.ProgressChanged event.
		/// </summary>
		/// <param name="percentProgress">The percentage, from 0 to 100, of the background operation that is complete.</param>
		/// <exception cref="System.InvalidOperationException">The WorkerReportsProgress property is set to false.</exception>
		public void ReportProgress(int percentProgress) {
			ReportProgress(percentProgress, null);
		}

		/// <summary>
		/// Starts execution of a background operation.
		/// </summary>
		/// <param name="argument">A parameter for use by the background operation to be executed in the BackgroundWorker.DoWork event handler.</param>
		/// <exception cref="System.InvalidOperationException">IsBusy is true.</exception>
		public void RunWorkerAsync(object argument) {
			lock (_syncRoot) {
				if (this.IsBusy) { throw new InvalidOperationException("IsBusy is true."); }
				this.IsBusy = true;
				this.CancellationPending = false;
			}
			ThreadPool.QueueUserWorkItem(new WaitCallback(Run), argument);
		}

		/// <summary>
		/// Starts execution of a background operation.
		/// </summary>
		/// <exception cref="System.InvalidOperationException">IsBusy is true.</exception>
		public void RunWorkerAsync() {
			RunWorkerAsync(null);
		}


		private bool _cancellationPending;

		/// <summary>
		/// Gets a value indicating whether the application has requested cancellation of a background operation.
		/// </summary>
		public bool CancellationPending {
			get {
				return this._cancellationPending;
			}
			private set {
				this._cancellationPending = value;
			}

		}


		private bool _isBusy;

		/// <summary>
		/// Gets a value indicating whether the BackgroundWorker is running an asynchronous operation.
		/// </summary>
		public bool IsBusy {
			get {
				lock (_syncRoot) {
					return this._isBusy;
				}
			}
			private set {
				lock (_syncRoot) {
					this._isBusy = value;
				}
			}
		}

		private bool _workerReportsProgress;
		/// <summary>
		/// Gets or sets a value indicating whether the BackgroundWorker can report progress updates.
		/// </summary>
		public bool WorkerReportsProgress {
			get {
				return this._workerReportsProgress;
			}
			set {
				this._workerReportsProgress = value;
			}
		}

		private bool _workerSupportsCancellation;
		/// <summary>
		/// Gets or sets a value indicating whether the BackgroundWorker supports asynchronous cancellation.
		/// </summary>
		public bool WorkerSupportsCancellation {
			get {
				return this._workerSupportsCancellation;
			}
			set {
				this._workerSupportsCancellation = value;
			}
		}

		/// <summary>
		/// Occurs when BackgroundWorker.RunWorkerAsync() is called.
		/// </summary>
		public event System.EventHandler<DoWorkEventArgs> DoWork;

		/// <summary>
		/// Occurs when BackgroundWorker.ReportProgress(System.Int32) is called.
		/// </summary>
		public event System.EventHandler<ProgressChangedEventArgs> ProgressChanged;

		/// <summary>
		/// Occurs when the background operation has completed, has been canceled, or has raised an exception.
		/// </summary>
		public event System.EventHandler<RunWorkerCompletedEventArgs> RunWorkerCompleted;

		private delegate void OnProgressChangedDelegate(ProgressChangedEventArgs e);
		private delegate void OnRunWorkerCompletedDelegate(RunWorkerCompletedEventArgs e);


		[System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Design", "CA1031:DoNotCatchGeneralExceptionTypes", Justification = "This is needed since this exception is forwarded further.")]
		private void Run(Object stateInfo) {
			try {
				DoWorkEventArgs e = new DoWorkEventArgs(stateInfo);
				this.OnDoWork(e);
				this.CancellationPending = false;
				this.IsBusy = false;
				if (e.Cancel) {
					_control.Invoke(new OnRunWorkerCompletedDelegate(OnRunWorkerCompleted), new RunWorkerCompletedEventArgs(null, null, true));
				} else {
					_control.Invoke(new OnRunWorkerCompletedDelegate(OnRunWorkerCompleted), new RunWorkerCompletedEventArgs(e.Result, null, false));
				}
			} catch (System.Exception ex) {
				_control.Invoke(new OnRunWorkerCompletedDelegate(OnRunWorkerCompleted), new RunWorkerCompletedEventArgs(null, ex, false));
			}
		}

	}





	/// <summary>
	/// Provides data for the BackgroundWorker.DoWork event handler.
	/// </summary>
	public class DoWorkEventArgs : System.ComponentModel.CancelEventArgs {

		/// <summary>
		/// Initializes new instance.
		/// </summary>
		/// <param name="argument">Specifies an argument for an asynchronous operation.</param>
		public DoWorkEventArgs(object argument)
			: base(false) {
			this.Argument = argument;
			this.Result = null;
		}

        private object _argument;
		/// <summary>
		/// Gets a value that represents the argument of an asynchronous operation.
		/// </summary>
        public object Argument { get { return this._argument; } private set { this._argument = value; } }

        private object _result;
		/// <summary>
		/// Gets or sets a value that represents the result of an asynchronous operation.
		/// </summary>
        public object Result { get { return this._result; } set { this._result = value; } }

	}





	/// <summary>
	/// Provides data for the BackgroundWorker.ProgressChanged event.
	/// </summary>
	public class ProgressChangedEventArgs : System.EventArgs {

		/// <summary>
		/// Initializes new instance.
		/// </summary>
		/// <param name="progressPercentage">The percentage of an asynchronous task that has been completed.</param>
		/// <param name="userState">A unique user state.</param>
		public ProgressChangedEventArgs(int progressPercentage, object userState) {
			this.ProgressPercentage = progressPercentage;
			this.UserState = userState;
		}

        private int _progressPercentage;
		/// <summary>
		/// Gets the asynchronous task progress percentage.
		/// </summary>
        public int ProgressPercentage { get { return this._progressPercentage; } private set { this._progressPercentage = value; } }

        private object _userState;
		/// <summary>
		/// Gets a unique user state.
		/// </summary>
        public object UserState { get { return this._userState; } private set { this._userState = value; } }

	}





	/// <summary>
	/// Provides data for the BackgroundWorker.RunWorkerCompleted event.
	/// </summary>
	public class RunWorkerCompletedEventArgs : System.EventArgs {

		/// <summary>
		/// Initializes new instance.
		/// </summary>
		/// <param name="result">The result of an asynchronous operation.</param>
		/// <param name="error">Any error that occurred during the asynchronous operation.</param>
		/// <param name="cancelled">A value indicating whether the asynchronous operation was canceled.</param>
		[System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Naming", "CA1726:UsePreferredTerms", MessageId = "cancelled", Justification = "Naming is same as Microsoft's.")]
		public RunWorkerCompletedEventArgs(object result, System.Exception error, bool cancelled) {
			this.Result = result;
			this.Error = error;
			this.Cancelled = cancelled;
		}


		private object _result;

		/// <summary>
		/// Gets a value that represents the result of an asynchronous operation.
		/// </summary>
		[System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Design", "CA1065:DoNotRaiseExceptionsInUnexpectedLocations", Justification = "Just following existing Microsoft's implementation.")]
		public object Result {
			get {
				if (this.Error != null) {
					throw new System.Reflection.TargetInvocationException(this.Error);
				}
				return this._result;
			}
			private set { this._result = value; }
		}

        private bool _cancelled;
		/// <summary>
		/// Gets a value indicating whether an asynchronous operation has been canceled.
		/// </summary>
		[System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Naming", "CA1726:UsePreferredTerms", MessageId = "Cancelled", Justification = "Naming is same as Microsoft's.")]
        public bool Cancelled { get { return this._cancelled; } private set { this._cancelled = value; } }

        private System.Exception _error;
		/// <summary>
		/// Gets a value indicating which error occurred during an asynchronous operation.
		/// </summary>
        public System.Exception Error { get { return this._error; } private set { this._error = value; } }

	}

}